/**
  * Copyright (c) Carnegie Mellon University.
  * See LICENSE.txt for the conditions of this license.
  */
/**
  * @note Code Review 2016-08-02 LemmaDB aspects only
  */
package edu.cmu.cs.ls.keymaerax.hydra

import java.io.FileOutputStream
import java.nio.channels.Channels

import edu.cmu.cs.ls.keymaerax.bellerophon.parser.{BelleParser, BellePrettyPrinter}
import edu.cmu.cs.ls.keymaerax.bellerophon.BelleExpr
import edu.cmu.cs.ls.keymaerax.core._
import edu.cmu.cs.ls.keymaerax.lemma._
import edu.cmu.cs.ls.keymaerax.parser.{KeYmaeraXArchiveParser, KeYmaeraXParser}
import edu.cmu.cs.ls.keymaerax.tools.ToolEvidence
import edu.cmu.cs.ls.keymaerax.core.{Formula, Sequent}
import edu.cmu.cs.ls.keymaerax.pt.{ElidingProvable, NoProof, ProvableSig, TermProvable}

import scala.collection.immutable.Nil
import scala.slick.driver.SQLiteDriver
import scala.util.Try

//import Tables.TacticonproofRow
import scala.slick.jdbc.StaticQuery.interpolation
import scala.slick.driver.SQLiteDriver.simple._

/**
 * Created by nfulton on 4/10/15.
 */
object SQLite {

  import Tables._

  lazy val ProdDB: SQLiteDB = new SQLiteDB(DBAbstractionObj.dblocation)
  /** Use this for all unit tests that work with the database, so tests don't infect the production database */
  lazy val TestDB: SQLiteDB = new SQLiteDB(DBAbstractionObj.testLocation)

  private val RULENAME_BRANCH_SEPARATOR: String = "@@"

  /** Stores lemmas in the Lemma table of the given database. */
  class UncachedSQLiteLemmaDB(db: SQLiteDB) extends LemmaDBBase {
    def readLemmas(ids: List[LemmaID]): Option[List[String]] = {
      db.getLemmas(ids.map(_.toInt))
    }

    override def get(ids: List[LemmaID]): Option[List[Lemma]] = {
      readLemmas(ids).map(_.map(Lemma.fromString)).map(_.map((lemma:Lemma) =>
        if (ProvableSig.PROOF_TERMS_ENABLED) {
          Lemma(TermProvable(ElidingProvable(lemma.fact.underlyingProvable), NoProof()), lemma.evidence, lemma.name)
        } else {
          lemma
        }
      ))
    }


    def writeLemma(id: LemmaID, lemma:String): Unit = {
      if(db.getLemma(id.toInt).nonEmpty)
        db.updateLemma(id.toInt, lemma)
      else {
        val createdId = db.createLemma()
        assert(createdId.toInt == id.toInt, s"Cannot create a new lemma with a custom ID unless that ID happens to be the next one generated by the SQLite database (requested: ${id}; created: ${createdId})")
        db.updateLemma(createdId, lemma)
      }
    }

    def createLemma():LemmaID = db.createLemma().toString

    override def remove(name: String): Unit = {
      db.deleteLemma(name.toInt)
    }

    override def removeAll(folder: String): Unit = ???

    override def deleteDatabase(): Unit = {
      db.deleteAllLemmas()
    }

    override def version(): String = db.getConfiguration("version").config("version")
  }

  def SQLiteLemmaDB (db: SQLiteDB) = new CachedLemmaDB(new UncachedSQLiteLemmaDB(db))

  class SQLiteDB(val dblocation: String) extends DBAbstraction {
    val sqldb: SQLiteDriver.backend.DatabaseDef = Database.forURL("jdbc:sqlite:" + dblocation, driver = "org.sqlite.JDBC")
    private val uncachedLemmaDB = new UncachedSQLiteLemmaDB(this)
    private val lemmaDB = SQLiteLemmaDB(this)
    private var currentSession:Session = null
    /* Statistics on the number of SQL operations performed in this session, useful for profiling. */
    private var nUpdates = 0
    private var nInserts = 0
    private var nSelects = 0

    implicit def session:Session = {
      if (currentSession == null || currentSession.conn.isClosed) {
        currentSession = sqldb.createSession()
        /* Enable write-ahead logging for SQLite - significantly improves write performance */
        sqlu"PRAGMA journal_mode = WAL".execute(currentSession)
        /* Enable foreign key support (off by default) */
        sqlu"PRAGMA foreign_keys = ON".execute(currentSession)
        /* Note: Setting synchronous = NORMAL introduces some risk of database corruption during power loss. According
         * to official documentation, that risk is less than the risk of the hard drive failing completely, but we
         * should at least be aware that the risk exists. Initial testing showed this to be about 8 times faster, so
         * it seems worth the risk. */
        sqlu"PRAGMA synchronous = NORMAL".execute(currentSession)
        sqlu"VACUUM".execute(currentSession)
      }
      currentSession
    }

    private def ensureExists(location: String): Unit = {
      if (!new java.io.File(location).exists()) {
        cleanup(location)
      }
    }
    ensureExists(DBAbstractionObj.dblocation)
    ensureExists(DBAbstractionObj.testLocation)

    /* withTransaction is not thread-safe if you reuse connections, it will raise an exception saying we're in autocommit
     * mode when we shouldn't be. So only ever call withTransaction inside a synchronized block. */
    def synchronizedTransaction[T](f: => T)(implicit session:Session): T =
      synchronized {
        session.withTransaction(f)
    }

    private[SQLite] def createLemma(): Int = {
      synchronizedTransaction({
        (Lemmas.map(_.lemma) returning Lemmas.map(_._Id.get))
          .insert(None)
      })
    }

    private[SQLite] def updateLemma(lemmaId: Int, lemma:String): Unit = {
      synchronizedTransaction({
        Lemmas.filter(_._Id === lemmaId).map(_.lemma).update(Some(lemma))
      })
    }

    private[SQLite] def getLemma(lemmaId: Int): Option[String] = {
      getLemmas(List(lemmaId)) match {
        case Some(lemmas) if lemmas.size <= 1 => lemmas.headOption
        case None => None
        case Some(l) if l.size>1 => throw new IllegalStateException("Should never get more than one lemma when asking for one lemma")
      }
    }

    private lazy val lemmaQuery = Compiled((lemmaId: Column[Int]) =>
      Lemmas.filter(_._Id === lemmaId).map(_.lemma.getOrElse("")))

    /** Allow retrieving lemmas in bulk to reduce the number of database queries */
    private[SQLite] def getLemmas(lemmaIds: List[Int]): Option[List[String]] = {
      synchronizedTransaction({
        if (lemmaIds.size > 1) {
          //@todo Code Review: This code should be revised to either select in SQL land from lemmaIds, or read all and filter, or read individual ones in a single transaction
          val lemmaMap = Lemmas.map{row => {
            val id = row._Id.getOrElse(throw new IllegalStateException("Did not expect lemma with null Id."))
            //Note: we allow the lemma text to be empty because of the Create / check we got ID / actually write lemma protocol assumed by CachedLemmaDB.
            //@todo Code Review: for lemmas that are in lemmaIds, assert not None, for all other lemmas, filter them out instead of converting to empty string
            val lemmaText = row.lemma.getOrElse("")
            (id, lemmaText)
          }}.list.toMap
          try {
            //@todo Code Review: check that lemmaIds really should not have "" names now
            Some(lemmaIds.map(lemmaMap(_)))
          } catch {
            case _:Exception => None
          }
        } else if (lemmaIds.size == 1) {
          Some(lemmaQuery(lemmaIds.head).list)
        } else Some(Nil)
      })
    }

    private[SQLite] def deleteLemma(lemmaId: Int): Unit = {
      synchronizedTransaction({
        // check that it deleted exactly one row
        val deletedEntries = Lemmas.filter(_._Id === lemmaId).delete
        assert(deletedEntries == 1, "deleting one identifier should delete one entry")
      })
    }

    private[SQLite] def deleteAllLemmas(): Unit = {
      synchronizedTransaction({
        Lemmas.delete
      })
    }

    override def syncDatabase():Unit = {
      sqlu"PRAGMA wal_checkpoint(FULL)".execute(session)
    }


    // Configuration and database setup and initialization and all DB communication

    override def getAllConfigurations: Set[ConfigurationPOJO] =
      synchronizedTransaction({
        nSelects = nSelects + 1
        Config.list.filter(_.configname.isDefined).map(_.configname.get).map(getConfiguration).toSet
      })

    private def blankOk(x: Option[String]): String = x match {
      case Some(y) => y
      case None => ""
    }

    override def getModelList(userId: String): List[ModelPOJO] = {
      synchronizedTransaction({
        nSelects = nSelects + 1
        Models.filter(_.userid === userId).list.map(element => new ModelPOJO(element._Id.get, element.userid.get,
          element.name.get, blankOk(element.date), blankOk(element.filecontents),
          blankOk(element.description), blankOk(element.publink), blankOk(element.title), element.tactic,
          getNumAllProofSteps(element._Id.get), element.istemporary.getOrElse(0) == 1))
      })
    }

    private def getNumAllProofSteps(modelId: Int): Int = {
      Proofs.innerJoin(Executionsteps).on((proofs, steps) => steps.proofid === proofs._Id).
        filter({ case (proofs, steps) => proofs.modelid === modelId }).length.run
    }

    private def configWithDefault(config: String, subconfig: String, default: Int): Int = {
      try {
        getConfiguration(config).config(subconfig).toInt
      } catch {case (_: NoSuchElementException) =>
        default
      }
    }
    override def createUser(username: String, password: String, mode: String): Unit = {
      /* Store passwords as a salted hash. Allow configuring number of iterations
       * since we may conceivably want to change it after deployment for performance reasons */
      val iterations = configWithDefault("security", "passwordHashIterations", 10000)
      val saltLength = configWithDefault("security", "passwordSaltLength", 512)
      val (hash, salt) = Password.generateKey(password, iterations, saltLength)
      synchronizedTransaction({
        Users.map(u => (u.email.get, u.hash.get, u.salt.get, u.iterations.get, u.level.get))
          .insert((username, hash, salt, iterations, Integer.parseInt(mode)))
        nInserts = nInserts + 1
      })}

    override def getUser(username: String): Option[UserPOJO] = synchronizedTransaction({
      nSelects = nSelects + 1
      val users =
        Users.filter(_.email === username)
          .list
          .map(m => new UserPOJO(m.email.get, m.level.get))
      if (users.length <= 1) users.headOption
      else throw new Exception("Primary keys aren't unique in models table.")
    })

    override def getTempUsers: List[UserPOJO] = synchronizedTransaction({
      nSelects = nSelects + 1
      Users.filter(_.level === 3).list.map(m => new UserPOJO(m.email.get, m.level.get))
    })

    /**
      * Poorly named -- either update the config, or else insert an existing key.
      * But in Mongo it was just update, because of the nested documents thing.
      *
      * @param config The new configuration.
      */
    override def updateConfiguration(config: ConfigurationPOJO): Unit =
      synchronizedTransaction({
        config.config.foreach(kvp => {
          val key = kvp._1
          val value = kvp._2
          nSelects = nSelects + 1
          val configExists = Config.filter(c => c.configname === config.name && c.key === key).list.nonEmpty

          if (configExists) {
            val q = for {l <- Config if l.configname === config.name && l.key === key} yield l.value
            q.update(Some(value))
            nUpdates = nUpdates + 1
          }
          else {
            Config.map(c => (c.configname.get, c.key.get, c.value.get))
              .insert((config.name, key, value))
            nInserts = nInserts + 1
          }
        })
      })

    private lazy val proofClosedQuery = Compiled((proofId: Column[Int]) =>
      Proofs.filter(p => p._Id === proofId && p.closed.getOrElse(0) === 1).exists)
    override def isProofClosed(proofId: Int): Boolean = synchronizedTransaction({
      proofClosedQuery(proofId).run
    })

    private val stepCountQuery = Compiled((proofId: Column[Int]) =>
      Executionsteps.filter(row =>
        row.proofid === proofId &&
        row.status === ExecutionStepStatus.toString(ExecutionStepStatus.Finished)).map(_._Id).countDistinct)

    //Proofs and Proof Nodes
    override def getProofInfo(proofId: Int): ProofPOJO =
      synchronizedTransaction({
        val stepCount = stepCountQuery(proofId).run
        nSelects = nSelects + 1
        val q = for { p <- Proofs if p._Id === proofId } yield (p.modelid, p.lemmaid)
        val (modelId, lemmaId) = q.run.head
        if (lemmaId.isEmpty) {
          val (lemmaId, _) = initializeProofForModel(modelId.get, None)
          q.update(modelId, Some(lemmaId))
        }

        val list = Proofs.filter(_._Id === proofId)
          .list
          .map(p => new ProofPOJO(p._Id.get, p.modelid, blankOk(p.name), blankOk(p.description),
            blankOk(p.date), stepCount, p.closed.getOrElse(0) == 1, p.lemmaid, p.istemporary.getOrElse(0) == 1, p.tactic))
        if (list.length > 1) throw new Exception("Duplicate proof " + proofId)
        else if (list.isEmpty) throw new Exception("Proof not found: " + proofId)
        else list.head
      })

    // Users
    override def userExists(username: String): Boolean =
      synchronizedTransaction({
        nSelects = nSelects + 1
        Users.filter(_.email === username).list.nonEmpty
      })


    override def getProofsForUser(userId: String): List[(ProofPOJO, String)] =
      synchronizedTransaction({
        val models = getModelList(userId)

        models.flatMap(model => {
          val modelName = model.name
          val proofs = getProofsForModel(model.modelId)
          proofs.map((_, modelName))
        })
      })

    private lazy val userOwnsProofQuery = Compiled((userId: Column[String], proofId: Column[Int]) => {
      (for {
        p <- Proofs.filter(_._Id === proofId)
        m <- Models.filter(_.userid === userId) if m._Id === p.modelid
      } yield p._Id).exists
    })

    override def userOwnsProof(userId: String, proofId: String): Boolean = synchronizedTransaction({
      userOwnsProofQuery(userId, proofId.toInt).run
    })

    override def checkPassword(username: String, password: String): Boolean =
      synchronizedTransaction({
        nSelects = nSelects + 1
        Users.filter(_.email === username).list.exists({row =>
          val hash = Password.hash(password.toCharArray, row.salt.get.getBytes("UTF-8"), row.iterations.get)
          Password.hashEquals(hash, row.hash.get)
        })
      })

    override def updateProofInfo(proof: ProofPOJO): Unit =
      synchronizedTransaction({
        nSelects = nSelects + 1
        Proofs.filter(_._Id === proof.proofId).update(proofPojoToRow(proof))
        nUpdates = nUpdates + 1
      })

    private def proofPojoToRow(p: ProofPOJO): ProofsRow =
      ProofsRow(_Id = Some(p.proofId), modelid = p.modelId, name = Some(p.name),
        description = Some(p.description), date =Some(p.date), closed = Some(if (p.closed) 1 else 0),
        lemmaid = p.provableId, istemporary = Some(if (p.temporary) 1 else 0), tactic = p.tactic
      )

    private def sqliteBoolToBoolean(x: Int) = if (x == 0) false else if (x == 1) true else throw new Exception()

    //returns id of create object
    override def getProofsForModel(modelId: Int): List[ProofPOJO] =
      synchronizedTransaction({
        nSelects = nSelects + 1
        Proofs.filter(_.modelid === modelId).list.map(p => {
          val stepCount = stepCountQuery(p._Id.get).run //@todo avoid ripple loading
          val closed: Boolean = sqliteBoolToBoolean(p.closed.getOrElse(0))
          val temporary: Boolean = sqliteBoolToBoolean(p.istemporary.getOrElse(0))
          ProofPOJO(p._Id.get, p.modelid, blankOk(p.name), blankOk(p.description), blankOk(p.date), stepCount,
            closed, p.lemmaid, temporary, p.tactic)
        })
      })

    override def deleteProofSteps(proofId: Int): Int = synchronizedTransaction({
      val deletedExecutionSteps = Executionsteps.filter(_.proofid === proofId).delete
      //@note deleting all steps, no need to update subgoal count
      val q = for { proof <- Proofs if proof._Id === proofId } yield (proof.closed, proof.lemmaid)
      val (closed, lemmaid) = q.run.head
      // delete associated lemma
      Lemmas.filter(_._Id === lemmaid).delete
      // reset closed flag and initial lemma
      q.update(Some(0), None)
      deletedExecutionSteps
    })

    override def deleteProof(proofId: Int): Boolean = synchronizedTransaction({
      deleteProofSteps(proofId)
      Proofs.filter(x => x._Id === proofId).delete == 1
    })

    //Models
    override def getUniqueModelName(userId: String, modelName: String): String = {
      val models = Models.filter(_.userid === userId).filter(_.name.startsWith(modelName)).map(_.name).list
      if (models.isEmpty) modelName
      else {
        val max = (0 :: models.filter(_.isDefined).map(s =>
          Try(s.get.substring(modelName.length).trim().stripPrefix("(").stripSuffix(")").toInt).toOption).
          filter(_.isDefined).map(_.get)).max
        s"$modelName (${max+1})"
      }
    }

    override def createModel(userId: String, name: String, fileContents: String, date: String,
                             description: Option[String] = None, publink: Option[String] = None,
                             title: Option[String] = None, tactic: Option[String] = None): Option[Int] =
      synchronizedTransaction({
        nSelects = nSelects + 1
        if (Models.filter(_.userid === userId).filter(_.name === name).list.isEmpty) {
          nInserts = nInserts + 1
          Some((Models.map(m => (m.userid.get, m.name.get, m.filecontents.get, m.date.get, m.description, m.publink, m.title, m.tactic))
            returning Models.map(_._Id.get))
            .insert(userId, name, fileContents, date, description, publink, title, tactic))
        }
        else None
      })

    override def updateModel(modelId: Int, name: String, title: Option[String], description: Option[String], content: Option[String]): Unit = synchronizedTransaction({
      assert(Models.filter(m => m._Id === modelId && m.filecontents === content).exists.run
        || Proofs.innerJoin(Executionsteps).on((proofs, steps) => steps.proofid === proofs._Id).
        filter({ case (proofs, _) => proofs.modelid === modelId }).length.run <= 0, "Updating model content only possible for models without proof steps")
      Models.filter(_._Id === modelId).map(m => (m.name, m.title, m.description, m.filecontents)).update(Some(name), title, description, content)
      nUpdates = nUpdates + 1
    })

    override def addModelTactic(modelId: String, fileContents: String): Option[Int] =
      synchronizedTransaction({
        nSelects = nSelects + 1
        val mId = Integer.parseInt(modelId)
        if (Models.filter(_._Id === mId).filter(_.tactic.isEmpty).list.isEmpty) {
          Some(Models.filter(_._Id === mId).map(_.tactic).update(Some(fileContents)))
        }
        else None
      })

    /** Initializes the proof by creating a provable from the model, returns the provable ID and optional substituted tactic. */
    private def initializeProofForModel(modelId: Int, tactic: Option[String]): (Int, Option[String]) = {
      val model = getModel(modelId)
      val entry = KeYmaeraXArchiveParser.parseProblem(model.keyFile, parseTactics=false)
      val d = entry.defs
      val problem = entry.model.asInstanceOf[Formula]

      val substTactic = tactic match {
        case None => None
        case Some(t) => Some(BellePrettyPrinter(BelleParser.parseWithInvGen(t, None, d)))
      }

      val provable = ProvableSig.startProof(problem)
      (createProvable(provable), substTactic)
    }

    override def createProofForModel(modelId: Int, name: String, description: String, date: String, tactic: Option[String]): Int =
      synchronizedTransaction({
        nInserts = nInserts + 2
        val (provableId, substTactic) = initializeProofForModel(modelId, tactic)
        val proofId =
          (Proofs.map(p => ( p.modelid.get, p.name.get, p.description.get, p.date.get, p.closed.get, p.lemmaid.get,
                             p.istemporary.get, p.tactic))
            returning Proofs.map(_._Id.get))
            .insert(modelId, name, description, date, 0, provableId, 0, substTactic)

        proofId
      })

    override def createProof(provable: ProvableSig): Int = synchronizedTransaction({
      nInserts = nInserts + 3
      val provableId = createProvable(provable)
      val proofId =
        (Proofs.map(p => ( p.closed.get, p.lemmaid.get, p.istemporary.get))
          returning Proofs.map(_._Id.get))
          .insert(0, provableId, 1)

//      val executableId = addBelleExpr(TactixLibrary.nil)
//      addExecutionStep(ExecutionStepPOJO(None, proofId, None, None, Some(0), None, 0,
//        ExecutionStepStatus.Finished, executableId, None, None, Some(provableId), userExecuted=false, "start"))

      proofId
    })

    override def getModel(modelId: Int): ModelPOJO =
      synchronizedTransaction({
        nSelects = nSelects + 1
        val models =
          Models.filter(_._Id === modelId)
            .list
            .map(m => new ModelPOJO(
              m._Id.get, m.userid.get, blankOk(m.name), blankOk(m.date), m.filecontents.get, blankOk(m.description),
              blankOk(m.publink), blankOk(m.title), m.tactic, getNumAllProofSteps(m._Id.get), m.istemporary.getOrElse(0) == 1
            ))
        if (models.length < 1) throw new Exception("getModel type should be an Option")
        else if (models.length == 1) models.head
        else throw new Exception("Primary keys aren't unique in models table.")
      })

    override def deleteModel(modelId: Int): Boolean =
      synchronizedTransaction({
        Models.filter(_._Id === modelId).delete
        Proofs.filter(_.modelid === modelId).list.map(prf => deleteProof(prf._Id.get))
        true
      })

    override def getConfiguration(configName: String): ConfigurationPOJO =
      synchronizedTransaction({
        nSelects = nSelects + 1
        val kvp = Config.filter(_.configname === configName)
          .filter(_.key.isDefined)
          .list
          .map(conf => (conf.key.get, blankOk(conf.value)))
          .toMap
        new ConfigurationPOJO(configName, kvp)
      })

    /**
      * Initializes a new database.
      */
    override def cleanup (): Unit = { cleanup(DBAbstractionObj.dblocation)}
    def cleanup(which: String): Unit = {
      val dbFile = this.getClass.getResourceAsStream("/keymaerax.sqlite")
      val target = new java.io.File(which)
      val targetStream = new FileOutputStream(target)
      targetStream.getChannel.transferFrom(Channels.newChannel(dbFile), 0, Long.MaxValue)
      targetStream.close()
      dbFile.close()
    }

    /** Deletes a provable and all associated sequents / formulas */
    override def deleteProvable(provableId: Int): Boolean = ???

    /**
      * Adds an execution step to an existing execution
      *
      * @note Implementations should enforce additional invarants -- never insert when branches or alt orderings overlap.
      */
    override def addExecutionStep(step: ExecutionStepPOJO): Int = {
      synchronizedTransaction({
        val status = ExecutionStepStatus.toString(step.status)
        val steps =
          Executionsteps.map(dbstep => (dbstep.proofid.get, dbstep.previousstep,
            dbstep.branchorder, dbstep.status.get, dbstep.executableid.get,
            dbstep.inputprovableid, dbstep.resultprovableid, dbstep.localprovableid, dbstep.userexecuted.get, dbstep.childrenrecorded.get,
            dbstep.rulename.get)
          ) returning Executionsteps.map(_._Id.get)
        val stepId = steps
            .insert((step.executionId, step.previousStep, step.branchOrder,
              status, step.executableId, step.inputProvableId, step.resultProvableId,
              step.localProvableId, step.userExecuted.toString, false.toString, step.ruleName))

        updateOpenSubgoalsCount(step.executionId, step.previousStep)

        nInserts = nInserts + 1
        stepId
      })
    }

    override def addAlternative(alternativeTo: Int, inputProvable: ProvableSig, trace:ExecutionTrace):Unit = {
//      def get(stepId: Int) = {
//        Executionsteps.filter(_._Id === stepId).list match {
//          case Nil => throw new Exception("Execution step not found")
//          case step :: _ => step
//        }
//      }
//      val oldStep = get(alternativeTo)
//      def addSteps(prev: Option[Int], globalProvable:ProvableSig, steps:List[ExecutionStep]): Unit = {
//        if (steps.nonEmpty) {
//          val thisStep = steps.head
//          val thisPOJO = get(thisStep.stepId)
//          val localProvable = getProvable(thisPOJO.localprovableid.get).provable
//          val outputProvable = globalProvable(localProvable, thisStep.branch)
//          val newStep = ExecutionStepPOJO(None, oldStep.executionid.get, prev, None, Some(thisStep.branch),
//            None, oldStep.alternativeorder.get + 1, ExecutionStepStatus.fromString(thisPOJO.status.get), thisPOJO.executableid.get,
//            None, None, thisPOJO.localprovableid, thisPOJO.userexecuted.get.toBoolean, thisPOJO.rulename.get, localProvable.subgoals.size)
//          val newId = addExecutionStep(newStep)
//          addSteps(Some(newId), outputProvable, steps.tail)
//        }
//      }
//      if(trace.steps.isEmpty) {
//        // Insert a null tactic with a higher alternative order
//        val nilExecutable = addBelleExpr(TactixLibrary.nil)
//        val localConclusion = inputProvable.subgoals(0)
//        // Generate a no-op local provable whose conclusion matches with the current state of the proof.
//        val localProvable = ProvableSig.startProof(localConclusion)
//        val newLocalProvableID = createProvable(localProvable)
//        val step = ExecutionStepPOJO(None, oldStep.executionid.get, oldStep.previousstep, None, Some(0), None,
//          oldStep.alternativeorder.get + 1, ExecutionStepStatus.Finished, nilExecutable, oldStep.inputprovableid,
//          oldStep.inputprovableid, Some(newLocalProvableID), userExecuted=false, "nil", localProvable.subgoals.size)
//        addExecutionStep(step)
//      } else {
//        addSteps(oldStep.previousstep, inputProvable, trace.steps)
//      }
    }

    /** Adds a Bellerophon expression as an executable and returns the new executableId */
    override def addBelleExpr(expr: BelleExpr): Int =
      synchronizedTransaction({
        val executableId =
          (Executables.map(_.belleexpr)
            returning Executables.map(_._Id.get))
          .insert(Some(BellePrettyPrinter(expr)))
        nInserts = nInserts + 1
        executableId
      })

    /** Stores a Provable in the database and returns its ID */
    override def createProvable(p: ProvableSig): Int = {
      synchronizedTransaction({
        val lemma = Lemma(p, Lemma.requiredEvidence(p, List(ToolEvidence(List("input" -> p.prettyString, "output" -> "true")))))
        lemmaDB.add(lemma).toInt
      })
    }

    /** Returns the executable with ID executableId */
    override def getExecutable(executableId: Int): ExecutablePOJO = {
      getExecutables(List(executableId)).head
    }

    /** Allow retrieving executables in bulk to reduce the number of database queries. */
    def getExecutables(executableIds: List[Int]): List[ExecutablePOJO] =
      synchronizedTransaction({
        nSelects = nSelects + 1
        val q = for {
          exe <- Executables
          if exe._Id inSetBind executableIds
        } yield (exe._Id.get, exe.belleexpr.get)
        val executableMap = q.run.map(exe => (exe._1, ExecutablePOJO(exe._1, exe._2))).toMap
        try {
          executableIds.map(executableMap)
        } catch {
          case _:Exception => throw new ProverException("getExecutable type should be an Option")
        }
      })

    /** Use escape hatch in prover core to create a new Provable */
    override def getProvable(lemmaId: Int): ProvablePOJO = {
      loadProvables(List(lemmaId)).head
    }

    def loadProvables(lemmaIds: List[Int]): List[ProvablePOJO] = {
      lemmaDB.get(lemmaIds.map(_.toString)) match {
        case None => throw new Exception (" No lemma for one of these IDs: " + lemmaIds)
        case Some(lemmas) => lemmas.zipWithIndex.map{case (lemma, id) => ProvablePOJO(id, lemma.fact)}
      }
    }



    private def tupleId(id: ProofTreeNodeId): (Option[Int], Option[Int]) = id match {
      case DbStepPathNodeId(step, branch) => (step, branch)
    }

    override def addAgendaItem(proofId: Int, initialProofNode: ProofTreeNodeId, displayName:String): Int = synchronizedTransaction({
      val (stepId, subgoalId) = tupleId(initialProofNode)
      (Agendaitems.map(item => (item.proofid.get, item.stepid, item.subgoalid, item.displayname.get))
        returning Agendaitems.map(_._Id.get))
        .insert(proofId, stepId, subgoalId, displayName)
    })

    override def updateAgendaItem(item:AgendaItemPOJO): Unit = synchronizedTransaction({
      val (stepId, subgoalId) = tupleId(item.initialProofNode)
      Agendaitems.filter(_._Id === item.itemId)
        .map(dbitem => (dbitem.proofid.get, dbitem.stepid, dbitem.subgoalid, dbitem.displayname.get))
        .update((item.proofId, stepId, subgoalId, item.displayName))
    })

    override def agendaItemsForProof(proofId: Int): List[AgendaItemPOJO] = {
      synchronizedTransaction({
        Agendaitems.filter(_.proofid === proofId)
        .list
        .map(item => AgendaItemPOJO(item._Id.get, item.proofid.get, DbStepPathNodeId(item.stepid, item.subgoalid), item.displayname.get))
      })
    }

    override def getAgendaItem(proofId: Int, initialProofNode: ProofTreeNodeId): Option[AgendaItemPOJO] = {
      val (stepId, subgoalId) = tupleId(initialProofNode)
      synchronizedTransaction({
        Agendaitems.filter{row => row.proofid === proofId && row.stepid === stepId && row.subgoalid === subgoalId}
          .list
          .map(item => AgendaItemPOJO(item._Id.get, item.proofid.get, DbStepPathNodeId(item.stepid, item.subgoalid), item.displayname.get))
          .headOption
      })
    }

    private def splitNameLabel(s: String): (String, Option[String]) = s.split(RULENAME_BRANCH_SEPARATOR).toList match {
      case rn :: Nil => (rn, None)
      case rn :: bl :: Nil => (rn, Some(bl))
    }

    /** Updates an executable step */
    override def updateExecutionStep(executionStepId: Int, step: ExecutionStepPOJO): Unit = {
      synchronizedTransaction({
        nSelects = nSelects + 1
        nUpdates = nUpdates + 1
        // update step
        Executionsteps.filter(_._Id === executionStepId)
        .map(dbstep => (dbstep.proofid.get, dbstep.previousstep,
          dbstep.branchorder, dbstep.status.get, dbstep.executableid.get,
          dbstep.inputprovableid, dbstep.resultprovableid, dbstep.localprovableid, dbstep.userexecuted.get, dbstep.childrenrecorded.get,
          dbstep.rulename.get, dbstep.numsubgoals, dbstep.numopensubgoals)
        ).update((step.executionId, step.previousStep, step.branchOrder,
          ExecutionStepStatus.toString(step.status), step.executableId, step.inputProvableId, step.resultProvableId,
          step.localProvableId, step.userExecuted.toString, false.toString,
          //@todo store branch labels separately
          step.ruleName + RULENAME_BRANCH_SEPARATOR + step.branchLabel.getOrElse(""),
          step.numSubgoals, step.numOpenSubgoals))

        // update parent's open subgoals
        updateOpenSubgoalsCount(step.executionId, step.previousStep)
      })
    }

    private lazy val stepQuery = Compiled((proofId: Column[Int], stepId: Column[Int]) =>
      Executionsteps.filter(row => row.proofid === proofId && row._Id === stepId).map(_.numsubgoals))

    private lazy val succStepCountQuery = Compiled((proofId: Column[Int], stepId: Column[Int]) =>
      Executionsteps.filter(row =>
        row.proofid === proofId &&
        row.previousstep === stepId &&
        row.status === ExecutionStepStatus.toString(ExecutionStepStatus.Finished)).map(_._Id).countDistinct)

    /** Recompute the open subgoals count of a step. */
    private def updateOpenSubgoalsCount(proofId: Int, stepId: Option[Int]): Unit = {
      stepId match {
        case None => // nothing to do
        case Some(sId) =>
          val succStepsCount = succStepCountQuery(proofId, sId).run
          val totalSubgoals = stepQuery(proofId, sId).list.head
          //@note select+update is faster than increment/decrement numSubGoals column
          Executionsteps.filter(_._Id === sId).map(_.numopensubgoals).update(totalSubgoals - succStepsCount)
      }
    }

    private lazy val stepParentQuery = Compiled((proofId: Column[Int], stepId: Column[Int]) =>
      Executionsteps.filter(row => row.proofid === proofId && row._Id === stepId).map(_.previousstep))

    /** Deletes execution steps. */
    override def deleteExecutionStep(proofId: Int, stepId: Int): Unit = {
      //val succStepsQuery = Executionsteps.filter(row => row.proofid === proofId && row._Id === stepId)
      val parent = stepParentQuery(proofId, stepId).list.head
      Executionsteps.filter(row => row.proofid === proofId && row._Id === stepId).delete
      updateOpenSubgoalsCount(proofId, parent)
    }

    def printStats(): Unit = {
      println("Updates: " + nUpdates + " Inserts: " + nInserts + " Selects: " + nSelects)
    }

    private lazy val executionStepsQuery = Compiled((proofId: Column[Int]) =>
      Executionsteps.filter(row => row.proofid === proofId &&
        row.status === ExecutionStepStatus.toString(ExecutionStepStatus.Finished)).
        sortBy(e => (e.previousstep.asc, e.branchorder.desc)))

    def proofSteps(executionId: Int): List[ExecutionStepPOJO] = {
      synchronizedTransaction({
        /* The Executionsteps table may contain many alternate histories for the same execution. In order to reconstruct
         * the current state of the world, we must pick the most recent alternative at every opportunity.*/
        var steps = executionStepsQuery(executionId).run
        val prevIds: scala.collection.mutable.Stack[Option[Int]] = new scala.collection.mutable.Stack()
        prevIds.push(None)
        var revResult: List[ExecutionStepPOJO] = Nil
        while (prevIds.nonEmpty) {
          val prevId = prevIds.top
          val (headSteps, tailSteps) = steps.partition(_.previousstep == prevId)
          if (headSteps.nonEmpty) {
            val head = headSteps.head
            val (ruleName: String, branchLabel: Option[String]) = splitNameLabel(head.rulename.get)
            revResult =
              ExecutionStepPOJO(head._Id, head.proofid.get, head.previousstep,
                head.branchorder, ExecutionStepStatus.fromString(head.status.get),
                head.executableid.get, head.inputprovableid, head.resultprovableid, head.localprovableid, head.userexecuted.get.toBoolean,
                ruleName, branchLabel, head.numsubgoals, head.numopensubgoals) :: revResult
            if (headSteps.tail.isEmpty) prevIds.pop
            prevIds.push(head._Id)
            steps = headSteps.tail ++ tailSteps
          } else {
            prevIds.pop
            steps = tailSteps
          }
        }
        revResult.reverse
      })
    }

    private lazy val executionStepsUnorderedQuery = (proofId: Int, stepIds: Set[Int]) =>
      Executionsteps.filter(row =>
        row.proofid === proofId &&
        (row._Id inSet stepIds) &&
        row.status === ExecutionStepStatus.toString(ExecutionStepStatus.Finished))

    private def proofStepsUnorderd(proofId: Int, stepIds: Set[Int]): List[ExecutionStepPOJO] = {
      executionStepsUnorderedQuery(proofId, stepIds).list.map(step => {
        val (ruleName: String, branchLabel: Option[String]) = splitNameLabel(step.rulename.get)
        ExecutionStepPOJO(step._Id, step.proofid.get, step.previousstep,
          step.branchorder, ExecutionStepStatus.fromString(step.status.get),
          step.executableid.get, step.inputprovableid, step.resultprovableid, step.localprovableid, step.userexecuted.get.toBoolean,
          ruleName, branchLabel, step.numsubgoals, step.numopensubgoals)
      })
    }

    private def getProofConclusion(proofId: Int): Sequent = {
      val proofInfo = getProofInfo(proofId)
      proofInfo.modelId match {
        case Some(modelId) =>
          val model = getModel(modelId)
          KeYmaeraXArchiveParser.parseAsProblemOrFormula(model.keyFile) match {
            case fml: Formula =>
              val sequent = Sequent(collection.immutable.IndexedSeq(), collection.immutable.IndexedSeq(fml))
              proofInfo.provableId match {
                case Some(pid) =>
                  val conclusion = getProvable(pid).provable.conclusion
                  assert(conclusion == sequent,
                    s"Model formula and associated provable's conclusion must match\nFormula:  $fml\nProvable: ${conclusion.succ.head}")
                  sequent
                case None => sequent
              }
            case _ => throw new Exception("Failed to parse model for proof " + proofId + " model " + modelId)
          }
        case None => proofInfo.provableId match {
          case Some(pid) => getProvable(pid).provable.conclusion
          case None => throw new Exception("Require either model or provable, but got neither")
        }
      }
    }

    /** Zips execution steps with auxiliary information (executable tactics, provables). */
    private def zipTrace(executionSteps: List[ExecutionStepPOJO],
                         provables: Option[(ProvableSig, Map[Option[Int], ProvableSig])]): List[ExecutionStep] = {
      executionSteps match {
        case step::steps =>
          val localProvable: () => ProvableSig = provables match {
            case None => () => getProvable(step.localProvableId.get).provable
            case Some((rootProvable, localProvables)) =>
              val localProvable = localProvables(step.stepId)
              val inputProvable = step.previousStep match {
                case prevId@Some(_) => localProvables(prevId)
                case None => rootProvable
              }
              assert(step.branchOrder < inputProvable.subgoals.size, "Open branch index " + step.branchOrder +
                " should point to a subgoal, but got only " + inputProvable.subgoals.size + " subgoals:\n" +
                inputProvable.subgoals.map(_.prettyString).mkString("\n"))
              assert(inputProvable.subgoals(step.branchOrder) == localProvable.conclusion,
                "Conclusion of subderivation " + localProvable.conclusion +
                  " should match open goal " + inputProvable.subgoals(step.branchOrder) +
                  " (subgoal " + step.branchOrder + ")")
              () => localProvable
          }
          val successorIds = steps.filter(_.previousStep == step.stepId).map(_.stepId.get)
          ExecutionStep(step.stepId.get, step.previousStep, step.executionId, localProvable,
            step.branchOrder, step.numSubgoals, step.numOpenSubgoals, successorIds, step.ruleName, step.executableId,
            step.userExecuted)  ::
            zipTrace(steps, provables)
        case Nil => Nil
        case _ => throw new ProverException("Bug in zipTrace")
      }
    }

    override def getExecutionSteps(executionId: Int): List[ExecutionStepPOJO] = proofSteps(executionId)

    private lazy val firstExecutionStepQuery = Compiled((proofId: Column[Int]) =>
      Executionsteps.filter(
        row => row.proofid === proofId &&
        row.status === ExecutionStepStatus.toString(ExecutionStepStatus.Finished) &&
        row.previousstep.isEmpty
      ))

    override def getFirstExecutionStep(proofId: Int): Option[ExecutionStepPOJO] =
      firstExecutionStepQuery(proofId).list.headOption.map(step => {
        val (ruleName: String, branchLabel: Option[String]) = splitNameLabel(step.rulename.get)
        ExecutionStepPOJO(step._Id, step.proofid.get, step.previousstep,
          step.branchorder, ExecutionStepStatus.fromString(step.status.get),
          step.executableid.get, step.inputprovableid, step.resultprovableid, step.localprovableid,
          step.userexecuted.get.toBoolean, ruleName, branchLabel, step.numsubgoals, step.numopensubgoals)
      })

    //@todo need better index to not degrade performance
//    private def openBranchingStepsSql(proofId: Int) =
//      sql"""SELECT exA._id
//            FROM executionSteps exA
//            LEFT OUTER JOIN executionSteps exB ON exA._id=exB.previousStep
//            GROUP BY exA._id
//            HAVING exA.proofId=$proofId AND COUNT(exB._id)<exA.numSubGoals""".as[Int]
//
//    private def closedBranchesSql(openSteps: List[Int]) =
//      sql"""SELECT previousStep,group_concat(branchOrder) FROM executionSteps WHERE previousStep IN (#${openSteps.mkString(",")}) GROUP BY previousStep""".as[(Int,String)]
//
//    override def getPlainOpenSteps(proofId: Int): List[(ExecutionStepPOJO,List[Int])] = synchronizedTransaction({
//      //trace.filter(parent => trace.count(s => parent.stepId == s.previousStep) < parent.numSubgoals)
//      def parseClosedBranches(closed: Option[String]): List[Int] = closed match {
//        case None => Nil
//        case Some(s) => s.split(",").map(_.toInt).toList
//      }
//
//      val openSteps = openBranchingStepsSql(proofId).list
//      val closedBranches = closedBranchesSql(openSteps).list.toMap
//      openSteps.map(s => getPlainExecutionStep(proofId, s).get ->
//        parseClosedBranches(closedBranches.get(s)))
//    })


    private lazy val openStepsQuery = Compiled((proofId: Column[Int]) =>
      Executionsteps.filter(row =>
        row.proofid === proofId &&
        row.numopensubgoals > 0 &&
        row.status === ExecutionStepStatus.toString(ExecutionStepStatus.Finished)))

    // performance degrades a little with increasing database volume
    private def closedBranchesSql(proofId: Int, openSteps: Set[Int]) =
      sql"""SELECT previousStep,group_concat(branchOrder) FROM executionSteps WHERE proofId=$proofId AND status='Finished' AND previousStep IN (#${openSteps.mkString(",")}) GROUP BY previousStep""".as[(Int,String)]

    // slightly slower than query above
//    private def closedBranchesSlick(proofId: Int, openSteps: Set[Int]) =
//      Executionsteps.filter(row =>
//        row.proofid === proofId &&
//        (row.previousstep inSet openSteps) &&
//        row.status === ExecutionStepStatus.toString(ExecutionStepStatus.Finished)
//      ).map(step => step.previousstep -> step.branchorder)

    override def getPlainOpenSteps(proofId: Int): List[(ExecutionStepPOJO,List[Int])] = synchronizedTransaction({
      val openSteps = openStepsQuery(proofId).list.map(step => {
        val (ruleName: String, branchLabel: Option[String]) = splitNameLabel(step.rulename.get)
        ExecutionStepPOJO(step._Id, step.proofid.get, step.previousstep,
          step.branchorder, ExecutionStepStatus.fromString(step.status.get),
          step.executableid.get, step.inputprovableid, step.resultprovableid, step.localprovableid,
          step.userexecuted.get.toBoolean, ruleName, branchLabel, step.numsubgoals, step.numopensubgoals)
      })
      val closedBranches = closedBranchesSql(proofId, openSteps.flatMap(_.stepId).toSet).list.toMap

      def parseClosedBranches(closed: Option[String]): List[Int] = closed match {
        case None => Nil
        case Some(s) => s.split(",").map(_.toInt).toList
      }

      openSteps.map(s => (s, parseClosedBranches(closedBranches.get(s.stepId.get))))
    })

    private lazy val executionStepQuery = Compiled((proofId: Column[Int], stepId: Column[Int]) =>
      Executionsteps.filter(row => row.proofid === proofId && row._Id === stepId &&
        row.status === ExecutionStepStatus.toString(ExecutionStepStatus.Finished)))

    override def getPlainExecutionStep(proofId: Int, stepId: Int): Option[ExecutionStepPOJO] = {
      executionStepQuery(proofId, stepId).run.headOption match {
        case None => None
        case Some(step) =>
          val (ruleName: String, branchLabel: Option[String]) = splitNameLabel(step.rulename.get)
          Some(ExecutionStepPOJO(step._Id, step.proofid.get, step.previousstep,
            step.branchorder, ExecutionStepStatus.fromString(step.status.get),
            step.executableid.get, step.inputprovableid, step.resultprovableid, step.localprovableid,
            step.userexecuted.get.toBoolean, ruleName, branchLabel, step.numsubgoals, step.numopensubgoals))
      }
    }

    private lazy val stepSuccessorsQuery = Compiled((proofId: Column[Int], prevStepId: Column[Int], branchOrder: Column[Int]) =>
      Executionsteps.filter(row =>
        row.proofid === proofId &&
        row.previousstep === prevStepId &&
        row.branchorder === branchOrder &&
        row.status === ExecutionStepStatus.toString(ExecutionStepStatus.Finished)))

    override def getPlainStepSuccessors(proofId: Int, prevStepId: Int, branchOrder: Int): List[ExecutionStepPOJO] = {
      stepSuccessorsQuery(proofId, prevStepId, branchOrder).run.map(step => {
        val (ruleName: String, branchLabel: Option[String]) = splitNameLabel(step.rulename.get)
        ExecutionStepPOJO(step._Id, step.proofid.get, step.previousstep,
          step.branchorder, ExecutionStepStatus.fromString(step.status.get),
          step.executableid.get, step.inputprovableid, step.resultprovableid, step.localprovableid,
          step.userexecuted.get.toBoolean, ruleName, branchLabel, step.numsubgoals, step.numopensubgoals)
      }).toList
    }

    override def getExecutionStep(proofId: Int, stepId: Int): Option[ExecutionStep] = getPlainExecutionStep(proofId, stepId) match {
      case None => None
      case Some(step) =>
        val localProvable = getProvable(step.localProvableId.get).provable
        val inputProvable = step.previousStep match {
          case None => ProvableSig.startProof(getProofConclusion(proofId))
          case Some(prevId) =>
            getProvable(getPlainExecutionStep(proofId, prevId).get.localProvableId.get).provable
        }
        //@todo extend getPlainExecutionStep to query for successor IDs right away
        val successorIds = getExecutionSteps(proofId).filter(_.previousStep == step.stepId).map(_.stepId.get)
        Some(ExecutionStep(step.stepId.get, step.previousStep, step.executionId, () => localProvable,
          step.branchOrder, step.numSubgoals, step.numOpenSubgoals, successorIds, step.ruleName, step.executableId,
          step.userExecuted))
    }

//    //@todo select last step of most recent history
//    override def getPlainLastExecutionStep(proofId: Int): Option[ExecutionStepPOJO] = getExecutionSteps(proofId).lastOption
//
//    override def getLastExecutionStep(proofId: Int): Option[ExecutionStep] = {
//      val steps = getExecutionSteps(proofId)
//      steps.lastOption match {
//        case None => None
//        case Some(step) =>
//          val localProvable = getProvable(step.localProvableId.get).provable
//          val inputProvable = step.previousStep match {
//            case None => ProvableSig.startProof(getProofConclusion(proofId))
//            case Some(prevId) => getProvable(steps.find(_.stepId == Some(prevId)).get.localProvableId.get).provable
//          }
//          Some(ExecutionStep(step.stepId.get, step.executionId, inputProvable, Some(localProvable), step.branchOrder.get,
//            step.alternativeOrder, step.ruleName, step.executableId, step.userExecuted))
//      }
//    }

    override def getStepProvable(proofId: Int, stepId: Option[Int], subgoal: Option[Int]): Option[ProvableSig] =
      (stepId, subgoal) match {
        case (None, None) => getProofInfo(proofId).provableId match {
          case None => None
          case Some(pId) => Some(getProvable(pId).provable)
        }
        case (Some(stId), None) => getPlainExecutionStep(proofId, stId) match {
          case Some(step) => step.localProvableId match {case None => None case Some(lpId) => Some(getProvable(lpId).provable)}
          case None => throw new IllegalArgumentException("Unknown step ID " + stId)
        }
        case (stId, Some(sgIdx)) => getStepProvable(proofId, stId, None) match {
          case None => None
          case Some(p) => Some(p.sub(sgIdx))
        }
      }

    override def getExecutionTrace(proofId: Int, withProvables: Boolean=true): ExecutionTrace = {
      /* This method has proven itself to be a resource hog, so this implementation attempts to minimize the number of
         DB calls. */
      val stepsCount = stepCountQuery(proofId).run
      if (stepsCount <= 0) {
        ExecutionTrace(proofId.toString, proofId.toString, Nil)
      } else {
        //@todo sqlite has a maximum of 999 variables in a query (could be changed, but any number might be too small)
        // -> may need to query steps and provables joined, but how to benefit from cached lemma DB then?
        val steps = proofSteps(proofId)
        val provables = if (withProvables) {
          val provableIds = steps.map(_.localProvableId.get)
          val provables = steps.map(_.stepId).zip(loadProvables(provableIds).map(_.provable)).toMap
          val conclusion = provables(steps.head.stepId).conclusion
          val initProvable = ProvableSig.startProof(conclusion)
          Some((initProvable, provables))
        } else None
        val traceSteps = zipTrace(steps, provables)
        ExecutionTrace(proofId.toString, proofId.toString, traceSteps)
      }
    }

    override def getInvariants(modelId: Int): Map[Expression, Formula] = {
      val model = getModel(modelId)
      var invariants: Map[Expression, Formula] = Map.empty
      KeYmaeraXParser.setAnnotationListener{case (program, formula) =>
        invariants = invariants.+((program, formula))
      }
      KeYmaeraXArchiveParser.parseAsProblemOrFormula(model.keyFile)
      invariants
    }

  }
}